# -*- coding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2017-05-02
"""

import os
import sys
import numpy
import struct
import socket
import logging
import datetime
import tempfile
import argparse
import threading

# Initialize Matplotlib backend
import matplotlib
matplotlib.use('Agg')

import SEAScope.lib
import SEAScope.lib.utils
import SEAScope.lib.plots

logger = logging.getLogger()
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)
logger.setLevel(logging.INFO)


def serialize_extraction():
    """"""
    logger.info('Serializeing extracted data...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%h%m%s'))
    output_path = os.path.join(output_dir, 'seascope_extract.pyo')

    granules = SEAScope.lib.get_extracted_data()
    SEAScope.lib.save_pyo(granules, output_path)

    logger.info('Done.')
    return True, '=> Serialization output saved in {}'.format(output_path)


def plot_extraction():
    """"""
    logger.info('Saving extracted data as PNGs...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%h%m%s'))

    granules = SEAScope.lib.get_extracted_data()

    for granule_uri in granules:
        granule_data = granules[granule_uri]['data']
        for field in granule_data:
            _data = numpy.flipud(granule_data[field])
            _uri = os.path.basename(granule_uri)

            output_file = '{}_{}.png'.format(_uri, field)
            output_path = os.path.join(output_dir, output_file)

            SEAScope.lib.plots.plot_2d_data(_uri, field, _data, output_path)

    return True, '=> PNGs generated for all granules in {}'.format(output_dir)


def plot_transect():
    """"""
    logger.info('Saving extracted data as PNGs...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%h%m%s'))

    granules = SEAScope.lib.get_extracted_data()

    if 'mask' not in granules:
        raise Exception('Missing mask in transects')

    dists = SEAScope.lib.get_dists(granules['mask'])

    transects = {}
    for granule_uri in granules:
        if 'mask' == granule_uri:
            continue

        granule = granules[granule_uri]
        for field in granule['data']:
            _uri = os.path.basename(granule_uri)
            uri = '{}: {}'.format(_uri, field)
            transects[uri] = granule['data'][field][0, :]

    output_file = '{}.png'.format('transects')
    output_path = os.path.join(output_dir, output_file)

    SEAScope.lib.plots.plot_transect(dists, transects, output_path)

    return True, '=> Result saved in {}'.format(output_path)


class ProcessingUnit(threading.Thread):
    """"""

    def __init__(self, s):
        """"""
        super(ProcessingUnit, self).__init__()
        self.socket = s

    def run(self):
        """"""
        buf = self.socket.recv(8)
        _opcode = struct.unpack('<Q', buf)
        opcode = _opcode[0]
        logger.debug('Received opcode {}'.format(opcode))
        if 0 == opcode:
            # noop
            pass
        elif 1 == opcode:
            ok, msg = plot_extraction()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        elif 2 == opcode:
            ok, msg = plot_transect()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        elif 3 == opcode:
            ok, msg = serialize_extraction()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        else:
            logger.error('Received unknown opcode {}'.format(opcode))

    def close(self):
        """"""
        self.socket.close()


def serve_forever():
    """"""
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(('127.0.0.1', 53450))
        s.listen(15)
    except socket.error:
        return False, 'Failed to create socket'

    logger.info('Processing server is running.')
    while True:
        try:
            client_socket, address = s.accept()
            processing_unit = ProcessingUnit(client_socket)
            processing_unit.start()
        except KeyboardInterrupt:
            return True, None

    return False, 'Something went wrong'


def seascope_processor():
    """"""
    ok, msg = serve_forever()
    if not ok:
        logger.error(msg)
        sys.exit(1)
