# -*- encoding: utf-8 -*-

"""
"""

import math
import logging
import datetime

logger = logging.getLogger(__name__)


def _id_generator(offset):
    """"""
    for value in range(offset, 100000):
        yield value

collection_id_generator = _id_generator(10)
granule_id_generator = _id_generator(1000)


def init_ids(collection_offset, granule_offset):
    """"""
    global collection_id_generator
    global granule_id_generator

    # collection IDs 1 and 2 are reserved for polygons and polylines
    if collection_offset < 3:
        collection_offset = 3
        logger.warning('Collection IDs 1 and 2 are reserved for internal use '
                       'in SEAScope. Your input has been modified to prevent '
                       'issues: collection_offset = 3')

    # Reserve 1000 granule IDs for the polygons and polylines the user will
    # draw in SEAScope
    if granule_offset < 1000:
        granule_offset = 1000
        logger.warning('Granule IDs below 1000 are reserved for internal use '
                       'in SEAScope. Your input has been modified to prevent '
                       'issues: granule_offset = 1000')

    collection_id_generator = _id_generator(collection_offset)
    granule_id_generator = _id_generator(granule_offset)


def create_collection(label):
    """"""
    unique_id = next(collection_id_generator)
    _rid = {'granuleLevel': False,
            'sourceId': 1,
            'collectionId': unique_id,
            'granuleId': 0,
            'variableId': 0}
    result = {'id': _rid,
              'mustBeCurrent': False,
              'xSeamless': False,
              'ySeamless': False,
              'label': label,
              'tags': {},
              'variables': [],
              'variablesList': {},
              'defaultVariable': 0}
    return unique_id, result


def create_variable(collection, label, fields, units='', dims=2):
    """"""
    if label in collection['variablesList']:
        logger.warn('Variable "{}" already exists for this collection and has '
                    'identifier "{}"'.format(
                    label, collection['variablesList'][label]))
        return None

    unique_id = len(collection['variablesList'].keys())
    collection['variablesList'][label] = unique_id

    rcfg = create_rendering_config(collection['id']['collectionId'], unique_id)
    r_method = 'RASTER'
    if 1 == dims:
        r_method = 'TRAJECTORIES'
    rcfg['renderMethod'] = r_method

    result = {'id': unique_id,
              'label': label,
              'units': '',
              'fields': fields,
              'defaultRenderingMethod': r_method,
              'tags': {},
              'collection': collection['id'],
              'rendering': rcfg}
    return result


def create_rendering_config(col_id=0, var_id=0):
    """"""
    target = {'granuleLevel': False,
              'sourceId': 1,
              'collectionId': col_id,
              'granuleId': 0,
              'variableId': var_id}

    result = {'rendered': True,
              'logscale': False,
              'min': 0,
              'max': 1,
              'opacity': 1.0,
              'zindex': 500,
              'color': [50, 12, 87],
              'colormap': 'grayscale',
              'renderMethod': 'RASTER',
              'particlesCount': 1000,
              'particleTTL': 10,
              'streamlineLength': 20,
              'streamlineSpeed': 0.0,
              'filterMode': 'NEAREST',
              'target': target}
    return result



def create_granule(collection_id, gcps, start, stop):
    """"""
    granule_id = next(granule_id_generator)
    _start = (start - datetime.datetime(1970, 1, 1)).total_seconds() * 1000
    _stop = (stop - datetime.datetime(1970, 1, 1)).total_seconds() * 1000
    data_id = 'user_generated_granule_{}'.format(granule_id)
    shape_arity = math.sqrt(len(gcps))
    shape = {'xArity': shape_arity,
             'yArity': shape_arity,
             'gcps': gcps}
    idf = {'uri': 'Python: {}'.format(data_id),
           'xArity': 0,
           'yArity': 0,
           'resolution': 1000000000,
           'subsampling_factor': 0,
           'shape': shape}
    metadata = {'sourceId': 1,
                'collectionId': collection_id,
                'granuleId': granule_id,
                'dataId': data_id,
                'dataModel': None,
                'start': int(_start),
                'stop': int(_stop),
                'uris': [idf],
                'title': data_id,
                'institution': '',
                'comment': '',
                'file_id': idf['uri'],
                'product_version': '0.0.0',
                'lat_min' : min([gcp['lat'] for gcp in gcps]),
                'lat_max' : max([gcp['lat'] for gcp in gcps]),
                'lon_min' : min([gcp['lon'] for gcp in gcps]),
                'lon_max' : max([gcp['lon'] for gcp in gcps]),
                'creator_email': '',
                'station_id': '',
                'platform': '',
                'sensor': ''}
    granule = {'id': {'granuleLevel': True,
                      'sourceId': 1,
                      'collectionId': collection_id,
                      'granuleId': granule_id,
                      'variableId': 0},
               'metadata': metadata,
               'data': {}}

    return granule_id, granule


def set_field(granule, field_name, field_data):
    """"""
    import numpy
    x_arity = 0
    y_arity = 0
    shape = numpy.shape(field_data)
    dims = len(shape)
    if 1 == dims:
        x_arity = shape[0]
        y_arity = 2
    elif 2 == dims:
        # Invert x and y since 2D data is transposed
        x_arity = shape[1]
        y_arity = shape[0]
    else:
        raise Exception('Field data must be either 1D or 2D')

    if granule['metadata']['dataModel'] is None:
        if 1 == dims:
            granule['metadata']['dataModel'] = 'TIME'
            granule['metadata']['uris'][0]['shape']['xArity'] = x_arity
            granule['metadata']['uris'][0]['shape']['yArity'] = 1
        elif 2 == dims:
            granule['metadata']['dataModel'] = 'ROW_CELL'
        granule['metadata']['uris'][0]['xArity'] = x_arity
        granule['metadata']['uris'][0]['yArity'] = y_arity

    # Check compatibility between field and granule geometries
    if (2 == dims) and ('TIME' == granule['metadata']['dataModel']):
        raise Exception('Adding a 2D field on a granule which already '
                        'contains a 1D field is not supported')
    elif (1 == dims) and ('TIME' != granule['metadata']['dataModel']):
        raise Exception('Adding a 2D field on a granule which already '
                        'contains a 1D field is not supported')

    ref_x_arity = granule['metadata']['uris'][0]['xArity']
    ref_y_arity = granule['metadata']['uris'][0]['yArity']
    if (x_arity != ref_x_arity) or (y_arity != ref_y_arity):
        raise Exception('All fields must have the same shape within a granule:'
                        'you can not add a field with a {}x{} shape in a '
                        'granule which contains {}x{} fields'.format(x_arity,
                        y_arity, ref_x_arity, ref_y_arity))

    # Find packing parameters
    vmin = numpy.min(field_data)
    vmax = numpy.max(field_data)
    offset = vmin
    scale = (vmax - vmin) / 254
    if vmin == vmax:
        scale = 1.0

    # Data must be flipped and transposed for SEAScope
    if 2 == dims:
        transposed = numpy.transpose(field_data)
    else:
        transposed = field_data[::-1]

    # Pack data as uint8 (max value <=> 254, 255 is reserved for mask)
    data_packed = (transposed - offset) / scale
    data_clipped = numpy.clip(data_packed, 0, 254).astype('uint8')

    # Mask invalid values (255 is used as fill value)
    fill_value = 255
    data_masked = numpy.ma.masked_invalid(data_clipped)
    data_masked[numpy.where(data_masked.mask)] = fill_value

    # Data must be stored as a 1D array (required for serialization)
    data_flattened = data_masked.flatten('F').tolist()

    channels = 1
    info = {'channels': channels,
            'xArity': x_arity,
            'yArity': y_arity,
            'dataType': 'uint8',
            'offsets': [offset],
            'scaleFactors': [scale],
            'fillValues': [fill_value]}

    granule['data'][field_name] = {'info': info,
                                   'buffer': data_flattened}


def get_lonlat(extraction, shape):
    """"""
    import numpy
    import scipy
    import scipy.interpolate

    if 2 != len(shape):
        raise Exception('The "shape" parameter must have exactly two '
                        'components.')

    gcps = extraction['meta']['gcps']
    i_gcp = numpy.array([_['i'] for _ in gcps])
    j_gcp = numpy.array([_['j'] for _ in gcps])
    lon_gcp = numpy.array([_['lon'] for _ in gcps])
    lat_gcp = numpy.array([_['lat'] for _ in gcps])

    # Enforce longitude continuity (to be improved)
    if (lon_gcp[-1] - lon_gcp[0]) > 180.0:
        logger.warn('Difference between first and last longitude exceeds '
                    '180 degrees, assuming IDL crossing and remapping '
                    'longitudes in [0, 360]')
        lon_gcp = numpy.mod((lon_gcp + 360.0), 360.0)

    # Restore shape of the GCPs
    gcps_shape = (4, 4)  # hardcoded in SEAScope
    i_shaped = numpy.reshape(i_gcp, gcps_shape)
    j_shaped = numpy.reshape(j_gcp, gcps_shape)
    lon_shaped = numpy.reshape(lon_gcp, gcps_shape)
    lat_shaped = numpy.reshape(lat_gcp, gcps_shape)

    # Build interpolation methods to get geographic coordinates based on
    # GCPs.
    interpolator = scipy.interpolate.RegularGridInterpolator
    regular_grid_elements = (j_shaped[:, 0], i_shaped[0, :])
    lon_interpolator = interpolator(regular_grid_elements, lon_shaped,
                                    method='linear', bounds_error=False,
                                    fill_value=None)
    lat_interpolator = interpolator(regular_grid_elements, lat_shaped,
                                    method='linear', bounds_error=False,
                                    fill_value=None)

    # Build the grid of interpolated i/j
    interp_i = numpy.arange(shape[1])
    interp_j = numpy.arange(shape[0])
    i_2D, j_2D = numpy.meshgrid(interp_i, interp_j)
    interp_ij = numpy.array((j_2D.flatten(), i_2D.flatten())).transpose()

    # Build the grid of interpolated lon/lat
    interp_lon = lon_interpolator(interp_ij)
    interp_lat = lat_interpolator(interp_ij)
    lon_2D = numpy.reshape(interp_lon, shape)
    lat_2D = numpy.reshape(interp_lat, shape)

    return lon_2D, lat_2D


def raw2data(raw_result):
    """"""
    import numpy
    clean_data = {}
    for extraction in raw_result:
        uri = extraction['uri'].decode('utf-8')
        if uri not in clean_data:
            clean_data[uri] = {}
            shape = (extraction['yArity'], extraction['xArity'])
            clean_data[uri]['data'] = {}
            clean_data[uri]['meta'] = {"shape": shape,
                                       "gcps": extraction['gcps'],
                                       'fill_values': [],
                                       'fields': [],
                                       'start': datetime.datetime.utcfromtimestamp(extraction['start']),
                                       'stop': datetime.datetime.utcfromtimestamp(extraction['stop'])}

        d = numpy.array(extraction['buffer'], dtype=numpy.float32)
        logger.debug('Full shape: {}'.format(d.shape))
        fill_values = extraction['fill_values']

        d0 = d[::4]
        data = [d0]

        if 1 < len(fill_values):
            d1 = d[1::4]
            data.append(d1)

        if 2 < len(fill_values):
            d2 = d[2::4]
            data.append(d2)

        for n in range(len(data)):
            if ((1 == len(data)) and
                (0 == len(extraction['fields'])) and
                ('mask' == uri)):
                field_name = 'mask'
                fill_value = 0
            else:
                field_name = (extraction['fields'][n]).decode('utf-8')
                fill_value = extraction['fill_values'][n]
            if field_name in clean_data[uri]['data']:
                # Field values already extracted from another variable
                continue
            dn_reshaped = numpy.reshape(data[n], shape)
            dn_masked = numpy.ma.masked_values(dn_reshaped, fill_value)
            dn_clean = numpy.ma.masked_invalid(dn_masked)
            clean_data[uri]['data'][field_name] = dn_clean
            clean_data[uri]['meta']['fill_values'].append(fill_value)
            clean_data[uri]['meta']['fields'].append(field_name)


    logger.info(' .:| Available granules |:.')
    logger.info(' --------------------------')
    for k in clean_data.keys():
        logger.info('• {}'.format(k))

    return clean_data
